
/* V 29.12.23 */
/* https://raw.githubusercontent.com/dbuezas/lgt8fx/master/package_lgt8fx_index.json */
/* Internal 32 MHz */

#include <EEPROM.h>
#include <OneWire.h>
#include <Wire.h>
#include <MsTimer2.h>
#include <Encoder.h>
#include <PID_v1.h>
#include <U8glib.h>
#include "GyverPWM.h"
#include "rus6x10.h",
#include "Rus_10x20.h",

U8GLIB_ST7920_128X64_4X u8g(13, 11, A4);

double Setpoint_F, Input_F, Output_F, Setpoint_C, Input_C, Output_C;
double Kp_F = 0.5, Ki_F = 3.0, Kd_F = 0.1;
double Kp_C = 0.5, Ki_C = 205.0, Kd_C = 0.5;
PID PID_FAN(&Input_F, &Output_F, &Setpoint_F, Kp_F, Ki_F, Kd_F, DIRECT);
PID PID_CUR(&Input_C, &Output_C, &Setpoint_C, Kp_C, Ki_C, Kd_C, DIRECT);
/*-----------------------------------------------------------------------------------------------------*/
Encoder myEnc(5, A5);
#define SB_Enc            4
long oldPosition        = 0;
byte SB_Touch           = 0;
/*-----------------------------------------------------------------------------------------------------*/
float Volt[]      = {0, 0};
float Volt_Lim[]  = {0, 0, 0, 8.45, 12.7, 16.95, 21.2, 25.35, 16.5, 14.0};
//                  {0, 0, 0, 8.4,  12.6, 16.8,  21.0, 25.2,  12.6, 12.7};
float Curr[]      = {0.0, 0.0, 1.5, 1.5, 4, 3};
float Curr_Var    = 0.239;
float Curr_End    = 0;
int   T_Curr;
const int Var_L   = 4094;
int   Tmp_V[]     = {Var_L, Var_L, 2004, 3784, 3444, 3114, 2834, 2534, 3114, 3114};
byte Zar          = 0;
byte Select       = 0;
int  Level        = Var_L;

byte Second[]     = {0};
byte Minute[]     = {0};
byte Hour[]       = {0}; 
int  Sec[]        = {0, 0, 0, 0, 0, 0, 0, 0, 0};
byte Dw           = 0;
int  Ind_C        = 300;
/*-----------------------------------------------------------------------------------------------------*/
OneWire DS_AK(7);                                     // Pin подключения датчика температуры аккумулятора
OneWire DS_BP(8);                                     // Pin подключения датчика температуры радиатора

#define POWER_MODE  0                                 // Режим питания, 0 - внешнее, 1 - паразитное
boolean FlagSensReady_AK;                             // Признак готовности данных с датчика
boolean FlagSensReady_BP;                             // Признак готовности данных с датчика
byte  bufData_BP[9];                                  // Буфер данных
byte  bufData_AK[9];                                  // Буфер данных
int   TimeCount;                                      // Счетчик времени измерения
byte  Present_AK          = 0;                        // Сигнал наличия термодатчика
byte  Present_BP          = 0;                        // Сигнал наличия термодатчика
byte  Flag_Cels           = 0;

float Cels_AK[]           = {0, 40};                  // Измеренная температура аккумулятора
byte  Cels_BP[]           = {0, 29, 30, 40, 50, 60, 70, 80, 90, 100};
byte  ii[]                = {0,  0, 30, 40, 50, 60, 70, 80, 90, 100};
byte  Cls[]               = {0, 0};
/*-----------------------------------------------------------------------------------------------------*/
const float R1              = 51000;                  // Резисторы делителя напряжения
const float R2              = 5100;
float V                     = 0.0;
/*-----------------------------------------------------------------------------------------------------*/
#define But_Zar               A3
#define PWM_OUT               9
#define Fan_PWM               6
#define Speed                 2
#define Beep                  12
byte  Beep_ON               = 0;
byte  Beep_OFF              = 0;
byte  Frequens[]            = {0, 0};
byte  Freq[]                = {0, 0, 0};
byte  Adress                = 4;
byte  Temp[5];
int   Tmp_L                 = 1;
byte  Flag_Plum[]           = {0, 0};
/*-----------------------------------------------------------------------------------------------------*/
#define CLR(x, y) (x &= (~(1 << y))) 
#define SET(x, y) (x |= (1 << y))
#define Zero_Time   Sec[0] = 0; Second[0] = 0; Minute[0] = 0; Hour[0] = 0;
/*=======================================================================================================
                                              Чтение EEPROM
=======================================================================================================*/
void EEPROM_Read()
{
  for (int i = 0; i <= Adress; i++)
  {
    Temp[i] = EEPROM.read(i);                // считываем значение по текущему адресу EEPROM
     Serial.print(" : ");Serial.print(i); Serial.print(" = "); Serial.print(Temp[i]);
  }
  Level = Temp[0] * 100 + Temp[1]; Curr[2] = (float)Temp[2] / 10; Curr[3] = (float)Temp[3] / 10; Curr[4] = (float)Temp[4] / 10;
  Serial.println(" << READ >> ");
}
/*=======================================================================================================
                                              Запись EEPROM
=======================================================================================================*/
void EEPROM_Write()
{
  Curr[2] *= 10; Curr[3] *= 10; Curr[4] *= 10;
  Temp[0] = Level / 100; Temp[1] = Level % 100; Temp[2] = Curr[2]; Temp[3] = Curr[3]; Temp[4] = Curr[4];
  for (int i = 0; i <= Adress; i++)
  {
    EEPROM.write(i, Temp[i]);
   Serial.print(" : ");  Serial.print(i); Serial.print(" = "); Serial.print(Temp[i]);
  }
  Serial.println(" << WRITE >> ");
}
/*=======================================================================================================
                                            Инициализация
=======================================================================================================*/
void setup()
{
  Serial.begin(115200);
  Wire.begin();

  PWM_resolution(9, 12, FAST_PWM);

  pinMode(PWM_OUT, OUTPUT);
  analogWrite(PWM_OUT, Var_L);

  EEPROM_Read();

  pinMode(A0, INPUT); 
  pinMode(A2, INPUT); 

  analogReadResolution(12);           // АЦП 12 БИТ
  analogReference(INTERNAL4V096);     // 4.096V

  pinMode(But_Zar, INPUT);
  pinMode(Speed, INPUT_PULLUP);
  pinMode(SB_Enc, INPUT_PULLUP);
  pinMode(Fan_PWM, OUTPUT);
  digitalWrite(Fan_PWM, HIGH);
  pinMode(Beep, OUTPUT);
  digitalWrite(Beep, LOW);
  Beep_ON = 1;
  
  u8g.begin();
  u8g.setColorIndex(7);                               // Определяем закрасить (1) или нет (0)
/*-----------------------------------------------------------------------------------------------------*/
  PID_FAN.SetTunings(Kp_F, Ki_F, Kd_F);
  PID_FAN.SetMode(AUTOMATIC);
  PID_FAN.SetOutputLimits(0, 255);  
  PID_FAN.Compute();
  PID_CUR.SetTunings(Kp_C, Ki_C, Kd_C);
  PID_CUR.SetMode(AUTOMATIC);
  PID_CUR.SetOutputLimits(0, Var_L);  
  PID_CUR.Compute();
/*-----------------------------------------------------------------------------------------------------*/
  attachInterrupt(0, ISR_F, FALLING);

  MsTimer2::set(1, Timing_ISR);                       // Задаем период прерывания по таймеру 1 мс
  MsTimer2::start();                                  // Разрешаем прерывание по таймеру

  if (Select >= 10) Select = 0;
}
/*=======================================================================================================
                                              Главный цикл
=======================================================================================================*/
void loop() 
{
  if (Flag_Cels == 1) Read_DS18B20();
  Read_Volt();
  if (Cels_AK[0] >= Cels_AK[1]) {Zar = 0; Select = 0; Level = Tmp_V[0]; Sec[3] = 0; Beep_ON = 1;}
/*-----------------------------------------------------------------------------------------------------*/
  for (byte i = 9; i >= 1; i--)
  {
    if (Cels_BP[0] <= Cels_BP[i]) Input_F = Cels_BP[0] + ii[i];
  }
  Setpoint_F = Frequens[0];
  PID_FAN.Compute();
  analogWrite(Fan_PWM, Output_F);
}
/*=======================================================================================================
                                              Encoder       // || - или, && - и
=======================================================================================================*/
void Encode()
{
  long newPosition = myEnc.read() / 2;
  if (oldPosition != newPosition) 
  {
    if (oldPosition < newPosition)
    {
      if (Select == 0) {if (Level <= Var_L - 10) Level += 20;}
      if ((Select >= 1) && (Select <= 7)) {if (Curr[2] >= 0.1) Curr[2] -= 0.1; Curr[3] = Curr[2]; Sec[5] = 0;}
      if (Select >= 8) {if (Curr[2] >= 0.1) Curr[2] -= 0.1; Curr[4] = Curr[2];}
    }
    if (oldPosition > newPosition)
    {
      if (Select == 0) {if (Level >= 1) Level -= 20;}
      if ((Select >= 1) && (Select <= 7)) {if (Curr[2] <= 10.0) Curr[2] += 0.1; Curr[3] = Curr[2]; Sec[5] = 0;}
      if (Select >= 8) {if (Curr[2] <= 10.0) Curr[2] += 0.1; Curr[4] = Curr[2];}
    }
    oldPosition = newPosition;
  }
/*-----------------------------------------------------------------------------------------------------*/
  if (digitalRead(SB_Enc) == 0)
  {
    digitalWrite(Beep, HIGH); Sec[3] = 0;
    switch (Select)
    {
      case 0: Select = 1; Curr[2] = Curr[3]; Sec[5] = 0; break;
      case 1: Select = 2; Curr[2] = Curr[5]; Level = Tmp_V[Select]; break;
      case 2: Select = 3; Curr[2] = Curr[3]; Level = Tmp_V[0]; Zero_Curr(); break;
      case 3: Select = 4; Curr[2] = Curr[3]; break;
      case 4: Select = 5; Curr[2] = Curr[3]; break;
      case 5: Select = 6; Curr[2] = Curr[3]; break;
      case 6: Select = 7; Curr[2] = Curr[3]; break;
      case 7: Select = 8; Curr[2] = Curr[4]; break;
      case 8: Select = 9; Curr[2] = Curr[3]; break;
      case 9: Select = 0; Curr[2] = Curr[3]; Level = Tmp_V[0]; Zar = 0; Zero_Time break;
    }

    SB_Touch = 0; Sec[2] = 0;
    while (SB_Touch == 0)
    {
      if (digitalRead(SB_Enc) == 0) SB_Touch = 0; else SB_Touch = 1;
      Sec[2]++;
      if (Sec[2] >= 10) digitalWrite(Beep, LOW);
      if (Sec[2] >= 180)
      {
        Sec[2] = 0; 
        EEPROM_Write(); EEPROM_Read();
        Sec[3] = 0; Beep_ON = 1;
        SB_Touch = 1;
      }
      delay(10);        
    }
  }
}
/*=======================================================================================================
                                        
=======================================================================================================*/
void ISR_F() {Frequens[1]++;}
void Zero_Curr()
{
  Input_C    = Curr[2];
  Setpoint_C = Curr[0];
  PID_CUR.Compute();
  Tmp_L = Level + Output_C;
  if (Tmp_L >= Var_L) Tmp_L = Var_L;
  if (Tmp_L <= 1) Tmp_L = 1;
  PWM_set(PWM_OUT, Tmp_L);
}
/*=======================================================================================================
                                         Чтение напряжения
=======================================================================================================*/
void Read_Volt() 
{
  Volt[1] = 0.0;
  Curr[1] = 0.0;
  byte  Del_V = 50;

  for (int v = 0; v < Del_V; v++) 
  {
    Volt[1] = Volt[1] + analogRead(0);
    Curr[1] = Curr[1] + analogRead(2);
    delay(10);
  }

  Volt[1] /= Del_V;
  V = (Volt[1] * 4.096) / 1023.0;
  Volt[0] = ((V / (R2 / (R1 + R2)))) / 4;

  Curr[1] /= Del_V; 
  Curr[0] = Curr[1] * (4.096 / 1023.0) * Curr_Var;
  if (Curr[0] <= 0.006) Curr[0] = 0;
}
/*=======================================================================================================
                                              Чтение температуры
=======================================================================================================*/
void Read_DS18B20()
{
  if (FlagSensReady_AK == true)
  {
    FlagSensReady_AK = false;                                 // данные готовы

    if (OneWire::crc8(bufData_AK, 8) == bufData_AK[8])      // проверка CRC данные правильные
    {
      Cels_AK[0] = ((float)((int)bufData_AK[0] | (((int)bufData_AK[1]) << 8)) * 0.0625 + 0.03125);
    }
    if (Present_AK == 0) Cels_AK[0] = 0;
  }
/*-----------------------------------------------------------------------------------------------------*/
  if (FlagSensReady_BP == true)
  {
    FlagSensReady_BP = false;                                  // данные готовы

    if (OneWire::crc8(bufData_BP, 8) == bufData_BP[8])        // проверка CRC данные правильные
    {
      Cels_BP[0] = ((float)((int)bufData_BP[0] | (((int)bufData_BP[1]) << 8)) * 0.0625 + 0.03125);
    }
  }
  Flag_Cels = 0;
}
/*=====================================================================================================*/
